# SPDX-License-Identifier: AGPL-3.0-or-later
"""Django forms for user management."""

import pwd
import re

from captcha.fields import CaptchaField
from django import forms
from django.contrib import auth, messages
from django.contrib.auth.forms import \
    AuthenticationForm as DjangoAuthenticationForm
from django.contrib.auth.forms import SetPasswordForm, UserCreationForm
from django.contrib.auth.hashers import check_password
from django.contrib.auth.models import Group, User
from django.core import validators
from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

import plinth.forms
import plinth.modules.ssh.privileged as ssh_privileged
from plinth.modules import first_boot
from plinth.utils import is_user_admin
from plinth.views import messages_error

from . import get_last_admin_user, privileged
from .components import UsersAndGroups


class AuthenticationForm(DjangoAuthenticationForm):
    """Authentication form with an additional username field attributes."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['username'].widget.attrs.update({
            'autofocus': 'autofocus',
            'autocapitalize': 'none',
            'autocomplete': 'username webauthn'
        })


class CaptchaForm(forms.Form):
    """Form with a CAPTCHA field to use after 3 invalid login attempts."""
    captcha = CaptchaField(
        label=_('Enter the letters in the image to proceed to the login page'))


class ValidNewUsernameCheckMixin:
    """Mixin to check if a username is valid for created new user."""

    def clean_username(self):
        """Check for username collisions with system users."""
        username = self.cleaned_data['username']
        if self.instance.username != username and \
                not self.is_valid_new_username():
            raise ValidationError(_('Username is taken or is reserved.'),
                                  code='invalid')

        return username

    def is_valid_new_username(self):
        """Check for username collisions with system users."""
        username = self.cleaned_data['username']
        existing_users = (a.pw_name.lower() for a in pwd.getpwall())
        if username.lower() in existing_users:
            return False

        if UsersAndGroups.is_username_reserved(username.lower()):
            return False

        return True


def _create_django_groups():
    """Ensure that all groups are present in the Django's group table."""
    group_choices = UsersAndGroups.get_group_choices()
    for group_name, _label in group_choices:
        Group.objects.get_or_create(name=group_name)

    return group_choices


class EmailFieldMixin:
    """Mixin to set common properties for the email field."""

    def __init__(self, *args, **kwargs):
        """Set basic properties for the email field."""
        super().__init__(*args, **kwargs)

        self.fields['email'].help_text = _(
            'Optional. Used to send emails to reset password and important '
            'notifications.')


class GroupsFieldMixin:
    """Mixin to set common properties for the group field."""

    def __init__(self, *args, **kwargs):
        """Set basic properties for the groups field.

        Also ensure that all the groups are created in django.
        """
        groups_dict = dict(_create_django_groups())

        super().__init__(*args, **kwargs)

        choices = []
        django_groups = sorted(self.fields['groups'].choices,
                               key=lambda choice: choice[1])
        for group_id, group_name in django_groups:
            try:
                group_id = group_id.value
            except AttributeError:
                pass

            # Show choices only from groups declared by apps.
            if group_name in groups_dict:
                label = groups_dict[group_name]
                if group_name == 'admin' and self.is_last_admin_user:
                    label = {'label': label, 'disabled': True}

                choices.append((group_id, label))

        self.fields['groups'].label = _('Permissions')
        self.fields['groups'].choices = choices
        self.fields['groups'].help_text = _(
            'Select which services should be available to the new '
            'user. The user will be able to log in to services that '
            'support single sign-on through LDAP, if they are in the '
            'appropriate group.<br /><br />Users in the admin group '
            'will be able to log in to all services. They can also '
            'log in to the system through SSH and have '
            'administrative privileges (sudo).')


@deconstructible
class UsernameValidator(validators.RegexValidator):
    """Username validator.

    Compared to django builtin ASCIIUsernameValidator, do not allow
    '+' characters and no '-' character at the beginning.

    """
    regex = r'^[\w.@][\w.@-]+\Z'
    message = gettext_lazy('Enter a valid username.')
    flags = re.ASCII


USERNAME_FIELD = forms.CharField(
    label=gettext_lazy('Username'), max_length=150,
    validators=[UsernameValidator()],
    help_text=gettext_lazy('Required. 150 characters or fewer. English '
                           'letters, digits and @/./-/_ only.'))


class PasswordConfirmForm(forms.Form):
    """Password confirmation form."""

    confirm_password = forms.CharField(
        widget=forms.PasswordInput,
        label=gettext_lazy('Authorization Password'))

    def __init__(self, *args, **kwargs):
        """Initialize form."""
        super().__init__(*args, **kwargs)

        self.fields['confirm_password'].help_text = _(
            'Enter the password for user "{user}" to authorize account '
            'modifications.').format(user=self.request.user.username)

    def clean_confirm_password(self):
        """Check that current user's password matches."""
        confirm_password = self.cleaned_data['confirm_password']
        password_matches = check_password(confirm_password,
                                          self.request.user.password)
        if not password_matches:
            raise ValidationError(_('Invalid password.'), code='invalid')

        return confirm_password


class CreateUserForm(ValidNewUsernameCheckMixin, EmailFieldMixin,
                     GroupsFieldMixin, plinth.forms.LanguageSelectionFormMixin,
                     PasswordConfirmForm, UserCreationForm):
    """Custom user create form.

    Include options to add user to groups.
    """

    username = USERNAME_FIELD

    language = plinth.forms.LanguageSelectionFormMixin.language

    class Meta(UserCreationForm.Meta):
        """Metadata to control automatic form building."""

        fields = ('username', 'email', 'password1', 'password2', 'groups',
                  'language', 'confirm_password')
        widgets = {
            'groups': plinth.forms.CheckboxSelectMultiple(),
        }

    def __init__(self, request, *args, **kwargs):
        """Initialize the form with extra request argument."""
        self.request = request
        self.is_last_admin_user = False
        super().__init__(*args, **kwargs)
        self.fields['username'].widget.attrs.update({
            'autofocus': 'autofocus',
            'autocapitalize': 'none',
            'autocomplete': 'username',
        })

    def save(self, commit=True):
        """Save the user model and create LDAP user if required."""
        user = super().save(commit)

        if commit:
            self.save_m2m()  # Django 3.x does not call save_m2m()

            user.userprofile.language = self.cleaned_data['language']
            user.userprofile.save()
            auth_username = self.request.user.username
            confirm_password = self.cleaned_data['confirm_password']

            try:
                privileged.create_user(user.get_username(),
                                       self.cleaned_data['password1'],
                                       auth_username, confirm_password)
            except Exception as error:
                messages.error(
                    self.request,
                    _('Creating LDAP user failed: {error}'.format(
                        error=error)))

            groups = user.groups.values_list('name', flat=True)
            for group in groups:
                try:
                    privileged.add_user_to_group(user.get_username(), group,
                                                 auth_username,
                                                 confirm_password)
                except Exception as error:
                    messages.error(
                        self.request,
                        _('Failed to add new user to {group} group: {error}').
                        format(group=group, error=error))

                group_object, created = Group.objects.get_or_create(name=group)
                group_object.user_set.add(user)

        return user


class UserUpdateForm(ValidNewUsernameCheckMixin, PasswordConfirmForm,
                     EmailFieldMixin, GroupsFieldMixin,
                     plinth.forms.LanguageSelectionFormMixin, forms.ModelForm):
    """When user info is changed, also updates LDAP user."""

    username = USERNAME_FIELD
    ssh_keys = forms.CharField(
        label=gettext_lazy('Authorized SSH Keys'), required=False,
        widget=forms.Textarea, help_text=gettext_lazy(
            'Setting an SSH public key will allow this user to '
            'securely log in to the system without using a '
            'password. You may enter multiple keys, one on each '
            'line. Blank lines and lines starting with # will be '
            'ignored.'))

    language = plinth.forms.LanguageSelectionFormMixin.language

    delete = forms.BooleanField(
        label=gettext_lazy('Delete user'), required=False,
        help_text=gettext_lazy(
            'Deleting the user account will also remove all the files '
            'related to the user. Deleting files can be avoided by '
            'setting the user account as inactive.'))

    class Meta:
        """Metadata to control automatic form building."""

        fields = ('username', 'email', 'groups', 'ssh_keys', 'language',
                  'is_active', 'delete', 'confirm_password')
        model = User
        widgets = {
            'groups': plinth.forms.CheckboxSelectMultipleWithReadOnly(),
        }

    def __init__(self, request, username, *args, **kwargs):
        """Initialize the form with extra request argument."""
        self.request = request
        self.username = username
        self.is_last_admin_user = get_last_admin_user() == self.username
        super().__init__(*args, **kwargs)
        self.fields['username'].widget.attrs.update({
            'autocapitalize': 'none',
            'autocomplete': 'username'
        })

        if not is_user_admin(request):
            self.fields['is_active'].widget = forms.HiddenInput()
            self.fields['groups'].disabled = True

        if self.is_last_admin_user:
            self.fields['is_active'].disabled = True
            self.fields['delete'].disabled = True

    def save(self, commit=True):
        """Update LDAP user name and groups after saving user model."""
        user = super().save(commit=False)
        # Profile is auto saved with user object
        user.userprofile.language = self.cleaned_data['language']
        auth_username = self.request.user.username
        confirm_password = self.cleaned_data['confirm_password']

        if commit:
            user.save()
            self.save_m2m()

            if self.cleaned_data.get('delete'):
                try:
                    # Remove system user
                    privileged.remove_user(user.get_username(), auth_username,
                                           confirm_password)
                except Exception as error:
                    messages_error(self.request, _('Failed to delete user.'),
                                   error)
                else:
                    # Remove Django user
                    user.delete()
                return user

            old_groups = privileged.get_user_groups(self.username)
            old_groups = [group for group in old_groups if group]

            if self.username != user.get_username():
                try:
                    privileged.rename_user(self.username, user.get_username())
                except Exception:
                    messages.error(self.request,
                                   _('Renaming LDAP user failed.'))

            new_groups = user.groups.values_list('name', flat=True)
            for old_group in old_groups:
                if old_group not in new_groups:
                    try:
                        privileged.remove_user_from_group(
                            user.get_username(), old_group, auth_username,
                            confirm_password)
                    except Exception:
                        messages.error(self.request,
                                       _('Failed to remove user from group.'))

            for new_group in new_groups:
                if new_group not in old_groups:
                    try:
                        privileged.add_user_to_group(user.get_username(),
                                                     new_group, auth_username,
                                                     confirm_password)
                    except Exception:
                        messages.error(self.request,
                                       _('Failed to add user to group.'))

            try:
                ssh_privileged.set_keys(user.get_username(),
                                        self.cleaned_data['ssh_keys'].strip(),
                                        auth_username, confirm_password)
            except Exception:
                messages.error(self.request, _('Unable to set SSH keys.'))

            is_active = self.cleaned_data['is_active']
            if self.initial['is_active'] != is_active:
                if is_active:
                    status = 'active'
                else:
                    status = 'inactive'
                try:
                    privileged.set_user_status(user.get_username(), status,
                                               auth_username, confirm_password)
                except Exception:
                    messages.error(self.request,
                                   _('Failed to change user status.'))

        return user

    def clean_groups(self):
        """Validate groups to ensure admin group for last admin.

        For the last admin user, we disable the checkbox for 'admin' group so
        that it can't be unchecked. However, this means that browser will no
        longer submit that value. Forcefully add 'admin' group in this case.

        """
        groups = self.cleaned_data['groups']
        if self.is_last_admin_user:
            groups = groups | self.fields['groups'].queryset.filter(
                **{'name': 'admin'})

        return groups


class UserChangePasswordForm(PasswordConfirmForm, SetPasswordForm):
    """Custom form that also updates password for LDAP users."""

    def __init__(self, request, *args, **kwargs):
        """Initialize the form with extra request argument."""
        self.request = request
        super().__init__(*args, **kwargs)
        self.fields['new_password1'].widget.attrs.update(
            {'autofocus': 'autofocus'})

    def save(self, commit=True):
        """Save the user model and change LDAP password as well."""
        user = super().save(commit)
        auth_username = self.request.user.username
        if commit:
            try:
                privileged.set_user_password(
                    user.get_username(), self.cleaned_data['new_password1'],
                    auth_username, self.cleaned_data['confirm_password'])
            except Exception:
                messages.error(self.request,
                               _('Changing LDAP user password failed.'))

        return user


class FirstBootForm(ValidNewUsernameCheckMixin, EmailFieldMixin,
                    auth.forms.UserCreationForm):
    """User module first boot step: create a new admin user."""

    username = USERNAME_FIELD

    class Meta(UserCreationForm.Meta):
        """Metadata to control automatic form building."""
        fields = ('username', 'email', 'password1')

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request')
        super().__init__(*args, **kwargs)
        # The wizard step has text to be read before entering the username.
        # Don't confuse screen readers by jumping directly to the username
        # field.
        self.fields['username'].widget.attrs.pop('autofocus', None)

    def save(self, commit=True):
        """Create and log the user in."""
        user = super().save(commit=commit)
        if commit:
            self.save_m2m()  # Django 3.x does not call save_m2m()

            first_boot.mark_step_done('users_firstboot')

            try:
                privileged.create_user(user.get_username(),
                                       self.cleaned_data['password1'])
            except Exception as error:
                messages.error(
                    self.request,
                    _('Creating LDAP user failed: {error}'.format(
                        error=error)))

            try:
                privileged.add_user_to_group(user.get_username(), 'admin')
            except Exception as error:
                messages.error(
                    self.request,
                    _('Failed to add new user to admin group: {error}'.format(
                        error=error)))

            _create_django_groups()

            admin_group = auth.models.Group.objects.get(name='admin')
            admin_group.user_set.add(user)

            self.login_user(self.cleaned_data['username'],
                            self.cleaned_data['password1'])

        return user

    def login_user(self, username, password):
        """Try to login the user with the credentials provided"""
        try:
            # Django axes requires the request attribute
            user = auth.authenticate(request=self.request, username=username,
                                     password=password)
            auth.login(self.request, user)
        except Exception:
            pass
        else:
            message = _('User account created, you are now logged in')
            messages.success(self.request, message)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Application manifest for roundcube."""

from django.utils.translation import gettext_lazy as _

clients = [{
    'name': _('Roundcube'),
    'platforms': [{
        'type': 'web',
        'url': '/roundcube'
    }]
}]

backup = {
    'data': {
        'files': [
            '/etc/roundcube/freedombox-config.php',
            '/var/lib/dbconfig-common/sqlite3/roundcube/roundcube'
        ]
    }
}

tags = [_('Email'), _('Contacts'), _('Web client')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure SOGo."""

import pathlib
import re
import shutil
import tempfile

from plinth import action_utils, utils
from plinth.actions import privileged
from plinth.db import postgres
from plinth.modules.email.privileged.domain import \
    get_domains as get_email_domains

DB_HOST = 'localhost'
DB_NAME = 'sogo_fbx'
DB_USER = 'sogo_fbx'
SERVICE_NAME = 'sogo'

DB_BACKUP_FILE = pathlib.Path('/var/lib/plinth/backups-data/sogo-database.sql')
CONFIG_FILE = pathlib.Path('/etc/sogo/sogo.conf')


@privileged
def setup() -> None:
    """Setup SOGo database and configuration."""
    database_password = utils.generate_password(16)
    postgres.create_database(DB_NAME, DB_USER, database_password)
    _create_config(database_password)


def _create_config(db_password: str):
    """Configure /etc/sogo/sogo.conf"""
    try:
        domain = _get_config_value('SOGoMailDomain')
        if not domain:
            # Try to get the domain configured for the email app
            domain = get_email_domains()['primary_domain']
    except FileNotFoundError:
        domain = 'localhost'

    connection = f'postgresql://{DB_USER}:{db_password}@{DB_HOST}/{DB_NAME}'

    content = f'''
{{
  /* General */
  SOGoMailDomain = "{domain}";
  SOGoLanguage = "English";
  SOGoTimeZone = "UTC";
  SOGoCalendarDefaultRoles = ("PublicViewer", "ConfidentialDAndTViewer");
  SOGoAppointmentSendEMailNotifications = YES;
  SOGoRefreshViewCheck = "every_minute";
  /* Allow users to add their own additional IMAP accounts */
  SOGoMailAuxiliaryUserAccountsEnabled = YES;
  /* Allow users to change their full name in default account */
  SOGoMailCustomFromEnabled = YES;

  /* Authentication */
  SOGoMaximumFailedLoginCount = "10";
  SOGoMaximumFailedLoginInterval = "300";
  SOGoFailedLoginBlockInterval = "300";

  /* Database */
  SOGoProfileURL = "{connection}/sogo_user_profile";
  OCSFolderInfoURL = "{connection}/sogo_folder_info";
  OCSSessionsFolderURL = "{connection}/sogo_sessions_folder";
  OCSEMailAlarmsFolderURL = "{connection}/sogo_alarms_folder";
  OCSStoreURL = "{connection}/sogo_store";
  OCSAclURL = "{connection}/sogo_acl";
  OCSCacheFolderURL = "{connection}/sogo_cache_folder";
  OCSAdminURL = "{connection}/sogo_admin";

  /* Cache */
  SOGoMemcachedHost = "127.0.0.1";

  /* SMTP */
  SOGoMailingMechanism = "smtp";
  SOGoSMTPServer = "smtp://127.0.0.1:587/?tls=YES&tlsVerifyMode=allowInsecureLocalhost";
  SOGoSMTPAuthenticationType = "PLAIN";

  /* IMAP */
  SOGoDraftsFolderName = "Drafts";
  SOGoSentFolderName = "Sent";
  SOGoTrashFolderName = "Trash";
  SOGoJunkFolderName = "Junk";
  SOGoIMAPServer = "imap://127.0.0.1:143/?tls=YES&tlsVerifyMode=allowInsecureLocalhost";
  SOGoSieveServer = "sieve://127.0.0.1:4190/?tls=YES&tlsVerifyMode=allowInsecureLocalhost";

  /* LDAP */
  SOGoUserSources = ({{
    type = "ldap";
    CNFieldName = "cn";
    IDFieldName = "uid";
    UIDFieldName = "uid";
    baseDN = "ou=users,dc=thisbox";
    canAuthenticate = YES;
    displayName = "Shared Addresses";
    hostname = "ldap://127.0.0.1:389";
    id = "directory";
    isAddressBook = YES;
  }});
}}'''  # noqa: E501
    CONFIG_FILE.touch(0o640, exist_ok=True)  # In case the file does not exist
    CONFIG_FILE.chmod(0o640)  # In case the file pre-existed
    shutil.chown(CONFIG_FILE, 'root', 'sogo')
    CONFIG_FILE.write_text(content, encoding='utf-8')


@privileged
def dump_database() -> None:
    """Dump database to file."""
    postgres.dump_database(DB_BACKUP_FILE, DB_NAME)


@privileged
def restore_database() -> None:
    """Restore database from file."""
    password = _read_db_password()
    postgres.restore_database(DB_BACKUP_FILE, DB_NAME, DB_USER, password)


def _read_db_password() -> str:
    """Extract the database password from /etc/sogo/sogo.conf using regex"""
    pattern = r'postgresql://[^:]+:([^@]+)@localhost'
    match = re.search(pattern, _get_config_value('SOGoProfileURL'))
    if not match:
        raise ValueError('Could not extract password')

    return match.group(1)


@privileged
def get_domain() -> str:
    """Get the value of SOGoMailDomain from /etc/sogo/sogo.conf"""
    return _get_config_value('SOGoMailDomain')


@privileged
def set_domain(domain: str):
    """Set the value of SOGoMailDomain in /etc/sogo/sogo.conf"""
    _set_config_value('SOGoMailDomain', domain)


def _get_config_value(key: str) -> str:
    """Return the value of a property from the configuration file."""
    process = action_utils.run(['plget', key], input=CONFIG_FILE.read_bytes(),
                               check=True)
    return process.stdout.decode().strip()


def _set_config_value(key: str, value: str):
    """Set the value of a property in the configuration file."""
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        temp_file.write(f'{{\n{key} = "{value}";\n}}'.encode('utf-8'))
        temp_file.close()
        action_utils.run(['plmerge', CONFIG_FILE, temp_file.name], check=True)
        pathlib.Path(temp_file.name).unlink()


@privileged
def uninstall() -> None:
    """Uninstall SOGo: drop database and configuration files."""
    postgres.drop_database(DB_NAME, DB_USER)
    CONFIG_FILE.unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure XMPP web client/jsxc."""

import logging

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages
from plinth.web_server import StaticFiles

from . import manifest

_description = [
    _('JSXC is a web client for XMPP. Typically it is used with an XMPP '
      'server running locally.'),
]

logger = logging.getLogger(__name__)


class JSXCApp(app_module.App):
    """FreedomBox app for JSXC."""

    app_id = 'jsxc'

    _version = 1

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('JSXC'), icon_filename='jsxc',
                               description=_description, manual_page='JSXC',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-jsxc', info.name, info.icon_filename,
                              info.tags, 'jsxc:index', parent_url_name='apps')
        self.add(menu_item)

        enable_state = app_module.EnableState('enable-state-jsxc')
        self.add(enable_state)

        shortcut = frontpage.Shortcut('shortcut-jsxc', name=info.name,
                                      icon=info.icon_filename,
                                      url=reverse_lazy('jsxc:jsxc'),
                                      clients=info.clients, tags=info.tags)
        self.add(shortcut)

        packages = Packages('packages-jsxc', [
            'libjs-jsxc', 'libjs-bootstrap4', 'libjs-jquery',
            'libjs-jquery-ui', 'node-jquery-slimscroll',
            'libjs-jquery-fullscreen'
        ])
        self.add(packages)

        firewall = Firewall('firewall-jsxc', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        directory_map = {
            '/static/jsxc/img': '/usr/share/libjs-jsxc/img/',
            '/static/jsxc/libjs-jsxc/lib': '/usr/share/javascript/jsxc/lib/',
            '/static/jsxc/libjs-jsxc/sound': '/usr/share/libjs-jsxc/sound/',
            '/static/jsxc/libjs-jsxc/': '/usr/share/libjs-jsxc/css/',
        }
        static_files = StaticFiles('static-files-jsxc',
                                   directory_map=directory_map)
        self.add(static_files)

        backup_restore = BackupRestore('backup-restore-jsxc',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        if not old_version:
            self.enable()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for avahi.
"""

from django.utils.translation import gettext_lazy as _

# Services that intend to make themselves discoverable will drop files into
# /etc/avahi/services. Currently, we don't intend to make that customizable.
# There is no necessity for backup and restore. This manifest will ensure that
# avahi enable/disable setting is preserved.
backup: dict = {}

tags = [_('Auto-discovery'), _('Local'), _('mDNS')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the sharing app.
"""

from django.urls import re_path

from .views import AddShareView, EditShareView, SharingAppView, remove

urlpatterns = [
    re_path(r'^apps/sharing/$', SharingAppView.as_view(), name='index'),
    re_path(r'^apps/sharing/add/$', AddShareView.as_view(), name='add'),
    re_path(r'^apps/sharing/(?P<name>[a-z0-9]+)/edit/$',
            EditShareView.as_view(), name='edit'),
    re_path(r'^apps/sharing/(?P<name>[a-z0-9]+)/remove/$', remove,
            name='remove'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox privacy app."""

from django import forms
from django.core import validators
from django.utils.translation import gettext_lazy as _

from plinth import cfg
from plinth.modules import names
from plinth.utils import format_lazy


class PrivacyForm(forms.Form):
    """Privacy configuration form."""

    help_ip_lookup_url = format_lazy(
        _('Optional Value. This URL is used to determine the publicly visible '
          'IP address of your {box_name}. The URL should simply return the '
          'IPv4 or IPv6 address where the client request comes from. Default '
          'is to use the service provided by the FreedomBox Foundation at '
          'https://ddns.freedombox.org/ip/. If empty, lookups are disabled '
          'and some functionality will fail.'), box_name=_(cfg.box_name))

    enable_popcon = forms.BooleanField(
        label=_('Periodically submit a list of apps used (suggested)'),
        required=False, help_text=format_lazy(
            _('Help Debian/{box_name} developers by participating in the '
              'Popularity Contest package survey program. When enabled, a '
              'list of apps used on this system will be anonymously submitted '
              'to Debian every week. Statistics for the data collected are '
              'publicly available at <a href="https://popcon.debian.org/" '
              'target="_blank">popcon.debian.org</a>. Submission happens over '
              'the Tor network for additional anonymity if Tor app is enabled.'
              ), box_name=_(cfg.box_name)))

    dns_fallback = forms.BooleanField(
        label=_('Allow using fallback DNS servers'), required=False,
        help_text=_(
            'Use well-known public DNS servers to resolve domain names in '
            'unusual circumstances where no DNS servers are known but '
            'internet connectivity is available. Can be disabled in most '
            'cases if network connectivity is stable and reliable.'))

    ip_lookup_url = forms.CharField(
        label=_('URL to look up public IP address'), required=False,
        help_text=help_ip_lookup_url,
        validators=[validators.URLValidator(schemes=['http', 'https'])])

    def __init__(self, *args, **kwargs):
        """Disable DNS fallback field if necessary."""
        super().__init__(*args, **kwargs)
        self.fields['dns_fallback'].disabled = (
            not names.is_resolved_installed())



# SPDX-License-Identifier: AGPL-3.0-or-later
"""URLs for the SOGo module."""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^apps/sogo/$', views.SOGoAppView.as_view(), name='index')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure Zoph web application."""

import logging

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import SharedDaemon
from plinth.modules.apache import privileged as apache_privileged
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages
from plinth.utils import format_lazy

from . import manifest, privileged

logger = logging.getLogger(__name__)

_description = [
    format_lazy(
        _('Zoph manages your photo collection. Photos are stored on your '
          '{box_name}, under your control. Instead of focusing on galleries '
          'for public display, Zoph focuses on managing them for your own '
          'use, organizing them by who took them, where they were taken, '
          'and who is in them. Photos can be linked to multiple hierarchical '
          'albums and categories. It is easy to find all photos containing a '
          'person, or photos taken on a date, or photos taken at a location '
          'using search, map and calendar views. Individual photos can be '
          'shared with others by sending a direct link.'),
        box_name=_(cfg.box_name)),
    format_lazy(
        _('The {box_name} user who setup Zoph will also become the '
          'administrator in Zoph. For additional users, accounts must be '
          'created both in {box_name} and in Zoph with the same user name.'),
        box_name=_(cfg.box_name))
]


class ZophApp(app_module.App):
    """FreedomBox app for Zoph."""

    app_id = 'zoph'

    _version = 3

    configure_when_disabled = False

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('Zoph'), icon_filename='zoph',
                               description=_description, manual_page='Zoph',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-zoph', info.name, info.icon_filename,
                              info.tags, 'zoph:index', parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-zoph', info.name,
                                      icon=info.icon_filename, url='/zoph/',
                                      clients=info.clients, tags=info.tags,
                                      login_required=True)
        self.add(shortcut)

        packages = Packages('packages-zoph', ['zoph', 'default-mysql-server'],
                            conflicts=['libpam-tmpdir'],
                            conflicts_action=Packages.ConflictsAction.REMOVE)
        self.add(packages)

        firewall = Firewall('firewall-zoph', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        dropin_configs = DropinConfigs('dropin-configs-zoph', [
            '/etc/apache2/conf-available/zoph-freedombox.conf',
        ])
        self.add(dropin_configs)

        webserver = Webserver('webserver-zoph', 'zoph',
                              urls=['https://{host}/zoph/'])
        self.add(webserver)

        webserver = Webserver('webserver-zoph-freedombox', 'zoph-freedombox')
        self.add(webserver)

        daemon = SharedDaemon('shared-daemon-zoph-mysql', 'mysql')
        self.add(daemon)

        backup_restore = ZophBackupRestore('backup-restore-zoph',
                                           **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        privileged.pre_install()
        with self.get_component('shared-daemon-zoph-mysql').ensure_running():
            # Database needs to be running for successful initialization or
            # upgrade of zoph database.
            super().setup(old_version)

        # Zoph brings mod-php as dependency and enables it. Re-run apache setup
        # to correct it. Disable mod-php, switch back to mpm-event, restart
        # apache2 if needed. Set old_version to an non-zero value so that
        # Snakeoil certificate is not regnerated.
        apache_privileged.setup(old_version=1)

        privileged.setup()
        if not old_version:
            self.enable()
        elif old_version < 2:
            if self.get_component('webserver-zoph').is_enabled():
                self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        # Before package uninstall, so that config file is still available
        privileged.uninstall()
        super().uninstall()


class ZophBackupRestore(BackupRestore):
    """Component to backup/restore Zoph database"""

    def backup_pre(self, packet):
        """Save database contents."""
        super().backup_pre(packet)
        privileged.dump_database()

    def restore_post(self, packet):
        """Restore database contents."""
        super().restore_post(packet)
        privileged.restore_database()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test various upgrade related utilities.
"""

from unittest.mock import patch

from .. import utils


def test_get_sources_list_codename(tmp_path):
    """Test retrieving codename from sources.list file."""
    list1 = '''
deb http://deb.debian.org/debian bookworm main non-free-firmware
deb-src http://deb.debian.org/debian bookworm main non-free-firmware

deb http://deb.debian.org/debian bookworm-updates main non-free-firmware
deb-src http://deb.debian.org/debian bookworm-updates main non-free-firmware

deb http://security.debian.org/debian-security/ bookworm-security main non-free-firmware
deb-src http://security.debian.org/debian-security/ bookworm-security main non-free-firmware
'''  # noqa: E501

    list2 = '''
deb http://deb.debian.org/debian stable main non-free-firmware
deb-src http://deb.debian.org/debian stable main non-free-firmware

deb http://deb.debian.org/debian bookworm-updates main non-free-firmware
deb-src http://deb.debian.org/debian bookworm-updates main non-free-firmware

deb http://security.debian.org/debian-security/ bookworm-security main non-free-firmware
deb-src http://security.debian.org/debian-security/ bookworm-security main non-free-firmware
'''  # noqa: E501

    list3 = '''
deb http://deb.debian.org/debian unstable main
deb http://deb.debian.org/debian trixie main
'''
    list4 = '''
deb http://deb.debian.org/debian sid main
deb http://deb.debian.org/debian trixie main
'''
    list5 = '''
deb http://deb.debian.org/debian testing main
deb http://deb.debian.org/debian trixie main
'''

    sources_list = tmp_path / 'sources.list'
    module = 'plinth.modules.upgrades.utils'
    with patch(f'{module}.sources_list', sources_list):
        sources_list.write_text(list1)
        assert utils.get_sources_list_codename() == 'bookworm'

        sources_list.write_text(list2)
        assert utils.get_sources_list_codename() is None

        sources_list.write_text(list3)
        assert utils.get_sources_list_codename() == 'unstable'

        sources_list.write_text(list4)
        assert utils.get_sources_list_codename() == 'unstable'

        sources_list.write_text(list5)
        assert utils.get_sources_list_codename() == 'testing'


@patch('subprocess.run')
def test_get_current_release(run):
    """Test that getting current release works."""
    run.return_value.stdout = b'test-release\ntest-codename\n\n'
    assert utils.get_current_release() == ('test-release', 'test-codename')


@patch('plinth.modules.upgrades.utils.get_sources_list_codename')
def test_is_distribution_unstable(get_sources_list_codename):
    """Test that checking for unstable distribution works."""
    get_sources_list_codename.return_value = 'unstable'
    assert utils.is_distribution_unstable()

    get_sources_list_codename.return_value = 'sid'
    assert utils.is_distribution_unstable()

    get_sources_list_codename.return_value = 'testing'
    assert not utils.is_distribution_unstable()


@patch('plinth.modules.upgrades.utils.get_current_release')
def test_is_distribution_rolling(get_current_release):
    """Test that checking for testing/unstable distribution works."""
    for value in ['unstable', 'testing', 'n/a']:
        get_current_release.return_value = (value, None)
        assert utils.is_distribution_rolling()

    get_current_release.return_value = ('forky', None)
    assert not utils.is_distribution_rolling()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Django form for configuring Gitweb.
"""

import re
from urllib.parse import urlparse

from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from django.utils.translation import gettext_lazy as _

from plinth.modules import gitweb

from . import privileged


def _get_branches(repo):
    """Get all the branches in the repository."""
    branch_data = privileged.get_branches(repo)
    default_branch = branch_data['default_branch']
    branches = branch_data['branches']

    if default_branch not in branches:
        branches.insert(0, default_branch)

    return [(branch, branch) for branch in branches]


def get_name_from_url(url):
    """Get a repository name from URL"""
    return urlparse(url).path.split('/')[-1]


def is_repo_url(url):
    """Check if URL is valid."""
    try:
        RepositoryValidator(input_should_be='url')(url)
    except ValidationError:
        return False

    return True


class RepositoryValidator:
    input_should_be = 'name'

    def __init__(self, input_should_be=None):
        if input_should_be is not None:
            self.input_should_be = input_should_be

    def __call__(self, value):
        """Validate that the input is a correct repository name or URL"""
        if self.input_should_be in ('url', 'url_or_name'):
            try:
                URLValidator(schemes=['http', 'https'],
                             message=_('Invalid repository URL.'))(value)
            except ValidationError:
                if self.input_should_be == 'url':
                    raise
            else:
                value = get_name_from_url(value)

        if (not re.match(r'^[a-zA-Z0-9-._]+$', value)) \
                or value.startswith(('-', '.')) \
                or value.endswith('.git.git'):
            raise ValidationError(_('Invalid repository name.'), 'invalid')


class CreateRepoForm(forms.Form):
    """Form to create and edit a new repository."""

    name = forms.CharField(
        label=_(
            'Name of a new repository or URL to import an existing repository.'
        ), strip=True,
        validators=[RepositoryValidator(input_should_be='url_or_name')],
        widget=forms.TextInput(attrs={'autocomplete': 'off'}))

    description = forms.CharField(
        label=_('Description of the repository'), strip=True, required=False,
        help_text=_('Optional, for displaying on Gitweb.'))

    owner = forms.CharField(label=_('Repository\'s owner name'), strip=True,
                            required=False,
                            help_text=_('Optional, for displaying on Gitweb.'))

    is_private = forms.BooleanField(
        label=_('Private repository'), required=False,
        help_text=_('Allow only authorized users to access this repository.'))

    def __init__(self, *args, **kwargs):
        """Initialize the form with extra request argument."""
        super().__init__(*args, **kwargs)
        self.fields['name'].widget.attrs.update({'autofocus': 'autofocus'})

    def clean_name(self):
        """Check if the name is valid."""
        name = self.cleaned_data['name']

        repo_name = name
        if is_repo_url(name):
            repo_name = get_name_from_url(name)

        if repo_name.endswith('.git'):
            repo_name = repo_name[:-4]

        for repo in gitweb.get_repo_list():
            if repo_name == repo['name']:
                raise ValidationError(
                    _('A repository with this name already exists.'))

        if is_repo_url(name):
            if not privileged.repo_exists(name):
                raise ValidationError('Remote repository is not available.')

        return name


class EditRepoForm(CreateRepoForm):
    """Form to create and edit a new repository."""

    name = forms.CharField(
        label=_('Name of the repository'),
        strip=True,
        validators=[RepositoryValidator()],
        help_text=_(
            'An alpha-numeric string that uniquely identifies a repository.'),
    )

    default_branch = forms.ChoiceField(
        label=_('Default branch'),
        help_text=_('Gitweb displays this as a default branch.'))

    def __init__(self, *args, **kwargs):
        """Initialize the form with extra request argument."""
        super().__init__(*args, **kwargs)
        branches = _get_branches(self.initial['name'])
        self.fields['default_branch'].choices = branches

    def clean_name(self):
        """Check if the name is valid."""
        name = self.cleaned_data['name']
        if 'name' in self.initial and name == self.initial['name']:
            return name

        if name.endswith('.git'):
            name = name[:-4]

        for repo in gitweb.get_repo_list():
            if name == repo['name']:
                raise ValidationError(
                    _('A repository with this name already exists.'))

        return name



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for basic system configuration
"""

from django import forms
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from plinth import cfg, frontpage
from plinth.modules.apache import get_users_with_website
from plinth.utils import format_lazy

from . import home_page_url2scid


def get_homepage_choices():
    """Return list of drop down choices for home page."""
    shortcuts = frontpage.Shortcut.list(web_apps_only=True)
    shortcut_choices = [(shortcut.component_id, shortcut.name)
                        for shortcut in shortcuts if shortcut.is_enabled()]
    uws_choices = \
        [(home_page_url2scid(url),
         format_lazy(gettext_lazy("{user}'s website"), user=user))
         for user, url in get_users_with_website().items()]
    apache_default = ('apache-default', _('Apache Default'))
    freedombox = ('freedombox', _('FreedomBox Service (Plinth)'))
    return [apache_default, freedombox] + uws_choices + shortcut_choices


class ConfigurationForm(forms.Form):
    """Main system configuration form"""

    homepage = forms.ChoiceField(
        label=gettext_lazy('Webserver Home Page'), help_text=format_lazy(
            gettext_lazy(
                'Choose the default page that must be served when '
                'someone visits your {box_name} on the web. A typical use '
                'case is to set your blog or wiki as the home page when '
                'someone visits the domain name. Note that once the home '
                'page is set to something other than {box_name} Service '
                '(Plinth), your users must explicitly type /freedombox to '
                'reach {box_name} Service (Plinth).'),
            box_name=gettext_lazy(cfg.box_name)), required=False,
        choices=get_homepage_choices)

    advanced_mode = forms.BooleanField(
        label=gettext_lazy('Show advanced apps and features'), required=False,
        help_text=gettext_lazy(
            'Show apps and features that require more technical '
            'knowledge.'))

    logging_mode = forms.ChoiceField(
        label=gettext_lazy('System-wide logging'),
        choices=(('none', gettext_lazy('Disable logging, for privacy')),
                 ('volatile',
                  gettext_lazy('Keep some in memory until a restart, '
                               'for performance')),
                 ('persistent',
                  gettext_lazy('Write to disk, useful for debugging'))),
        help_text=gettext_lazy(
            'Logs contain information about who accessed the system and debug '
            'information from various services'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Shutdown/restart the system."""

from plinth import action_utils
from plinth.actions import privileged


@privileged
def restart():
    """Restart the system."""
    action_utils.run('reboot', check=False)


@privileged
def shutdown():
    """Shut down the system."""
    action_utils.run(['shutdown', 'now'], check=False)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for storage.
"""
from django.utils.translation import gettext_lazy as _

backup: dict = {}

tags = [_('Disks'), _('Usage'), _('Auto-mount'), _('Expand partition')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for gitweb views.
"""

from unittest.mock import Mock, patch

import pytest
from django import urls
from django.contrib.messages.storage.fallback import FallbackStorage
from django.http.response import Http404

from plinth import module_loader
from plinth.modules.gitweb import views

# For all tests, use plinth.urls instead of urls configured for testing
pytestmark = pytest.mark.urls('plinth.urls')

EXISTING_REPOS = [
    {
        'name': 'something',
        'description': '',
        'owner': '',
        'access': 'public',
        'is_private': False,
        'default_branch': 'main',
    },
    {
        'name': 'something2',
        'description': '',
        'owner': '',
        'access': 'private',
        'is_private': True,
        'default_branch': 'main',
    },
]


@pytest.fixture(autouse=True, scope='module')
def fixture_gitweb_urls():
    """Make sure gitweb app's URLs are part of plinth.urls."""
    with patch('plinth.module_loader._modules_to_load', new=[]) as modules, \
            patch('plinth.urls.urlpatterns', new=[]):
        modules.append('plinth.modules.gitweb')
        module_loader.include_urls()
        yield


@pytest.fixture(autouse=True)
def gitweb_patch():
    """Patch gitweb."""
    privileged = 'plinth.modules.gitweb.privileged'
    with patch('plinth.modules.gitweb.get_repo_list') as get_repo_list, \
         patch('plinth.app.App.get') as app_get, \
         patch(f'{privileged}.create_repo'), \
         patch(f'{privileged}.repo_exists') as repo_exists, \
         patch(f'{privileged}.repo_info') as repo_info, \
         patch(f'{privileged}.rename_repo'), \
         patch(f'{privileged}.set_repo_description'), \
         patch(f'{privileged}.set_repo_owner'), \
         patch(f'{privileged}.set_repo_access'), \
         patch(f'{privileged}.set_default_branch'), \
         patch(f'{privileged}.delete_repo'), \
         patch(f'{privileged}.get_branches') as get_branches:
        repo_exists.return_value = True
        repo_info.return_value = dict(EXISTING_REPOS[0])
        get_branches.return_value = {
            'default_branch': 'main',
            'branches': ['main', 'branch1']
        }
        get_repo_list.return_value = [{
            'name': EXISTING_REPOS[0]['name']
        }, {
            'name': EXISTING_REPOS[1]['name']
        }]
        app = Mock()
        app_get.return_value = app
        app.update_service_access.return_value = None

        yield


def make_request(request, view, **kwargs):
    """Make request with a message storage."""
    setattr(request, 'session', 'session')
    messages = FallbackStorage(request)
    setattr(request, '_messages', messages)
    response = view(request, **kwargs)

    return response, messages


def test_repos_view(rf):
    """Test that a repo list has correct view data."""
    with patch('plinth.views.AppView.get_context_data',
               return_value={'is_enabled': True}):
        view = views.GitwebAppView.as_view()
        response, _ = make_request(rf.get(''), view)

        assert response.context_data['repos'] == [{
            'name': EXISTING_REPOS[0]['name']
        }, {
            'name': EXISTING_REPOS[1]['name']
        }]
        assert response.status_code == 200


def test_create_repo_view(rf):
    """Test that repo create view sends correct success message."""
    form_data = {
        'gitweb-name': 'something_other',
        'gitweb-description': 'test description',
        'gitweb-owner': 'test owner',
        'gitweb-is_private': True
    }
    request = rf.post(urls.reverse('gitweb:create'), data=form_data)
    view = views.CreateRepoView.as_view()
    response, messages = make_request(request, view)

    assert list(messages)[0].message == 'Repository created.'
    assert response.status_code == 302


def test_create_repo_duplicate_name_view(rf):
    """Test that repo create view shows correct error message."""
    form_data = {
        'gitweb-name': EXISTING_REPOS[0]['name'],
        'gitweb-description': '',
        'gitweb-owner': ''
    }
    request = rf.post(urls.reverse('gitweb:create'), data=form_data)
    view = views.CreateRepoView.as_view()
    response, _ = make_request(request, view)

    assert response.context_data['form'].errors['name'][
        0] == 'A repository with this name already exists.'
    assert response.status_code == 200


def test_create_repo_invalid_name_view(rf):
    """Test that repo create view shows correct error message."""
    form_data = {
        'gitweb-name': '.something_other',
        'gitweb-description': '',
        'gitweb-owner': ''
    }
    request = rf.post(urls.reverse('gitweb:create'), data=form_data)
    view = views.CreateRepoView.as_view()
    response, _ = make_request(request, view)

    assert response.context_data['form'].errors['name'][
        0] == 'Invalid repository name.'
    assert response.status_code == 200


def test_create_repo_failed_view(rf):
    """Test that repo creation failure sends correct error message."""
    general_error_message = "An error occurred while creating the repository."
    error_description = 'some error'
    with patch('plinth.modules.gitweb.create_repo',
               side_effect=PermissionError(error_description)):
        form_data = {
            'gitweb-name': 'something_other',
            'gitweb-description': '',
            'gitweb-owner': ''
        }
        request = rf.post(urls.reverse('gitweb:create'), data=form_data)
        view = views.CreateRepoView.as_view()
        response, messages = make_request(request, view)

        assert list(messages)[0].message == '{0} {1}'.format(
            general_error_message, error_description)
        assert response.status_code == 302


def test_clone_repo_view(rf):
    """Test that cloning repo sends correct succcess message."""
    form_data = {
        'gitweb-name': 'https://example.com/test.git',
        'gitweb-description': '',
        'gitweb-owner': ''
    }
    request = rf.post(urls.reverse('gitweb:create'), data=form_data)
    view = views.CreateRepoView.as_view()
    response, messages = make_request(request, view)

    with pytest.raises(AttributeError):
        getattr(response.context_data['form'], 'errors')

    assert list(messages)[0].message == 'Repository created.'
    assert response.status_code == 302


def test_clone_repo_missing_remote_view(rf):
    """Test that cloning non-existing repo shows correct error message."""
    with patch('plinth.modules.gitweb.privileged.repo_exists',
               return_value=False):
        form_data = {
            'gitweb-name': 'https://example.com/test.git',
            'gitweb-description': '',
            'gitweb-owner': ''
        }
        request = rf.post(urls.reverse('gitweb:create'), data=form_data)
        view = views.CreateRepoView.as_view()
        response, _ = make_request(request, view)

        assert response.context_data['form'].errors['name'][
            0] == 'Remote repository is not available.'
        assert response.status_code == 200


def test_edit_repository_view(rf):
    """Test that editing repo sends correct success message."""
    form_data = {
        'gitweb-name': 'something_other.git',
        'gitweb-description': 'test-description',
        'gitweb-owner': 'test-owner',
        'gitweb-is_private': True,
        'gitweb-default_branch': 'branch1',
    }
    url = urls.reverse('gitweb:edit',
                       kwargs={'name': EXISTING_REPOS[0]['name']})
    request = rf.post(url, data=form_data)
    view = views.EditRepoView.as_view()
    response, messages = make_request(request, view,
                                      name=EXISTING_REPOS[0]['name'])

    assert list(messages)[0].message == 'Repository edited.'
    assert response.status_code == 302


def test_edit_nonexisting_repository_view(rf):
    """Test that trying to edit non-existing repository raises 404."""
    non_existing_repo = {'name': 'something_other'}
    request = rf.post(urls.reverse('gitweb:edit', kwargs=non_existing_repo),
                      data={})
    view = views.EditRepoView.as_view()

    with pytest.raises(Http404):
        make_request(request, view, **non_existing_repo)


def test_edit_repository_duplicate_name_view(rf):
    """Test that renaming to already existing repo name shows correct error
    message.

    """
    form_data = {
        'gitweb-name': EXISTING_REPOS[1]['name'],
        'gitweb-description': 'test-description',
        'gitweb-owner': 'test-owner'
    }
    url = urls.reverse('gitweb:edit',
                       kwargs={'name': EXISTING_REPOS[0]['name']})
    request = rf.post(url, data=form_data)
    view = views.EditRepoView.as_view()
    response, _ = make_request(request, view, **EXISTING_REPOS[0])

    assert response.context_data['form'].errors['name'][
        0] == 'A repository with this name already exists.'
    assert response.status_code == 200


def test_edit_repository_invalid_name_view(rf):
    """Test that renaming repo to invalid name shows correct error message."""
    form_data = {
        'gitweb-name': '.something',
        'gitweb-description': 'test-description',
        'gitweb-owner': 'test-owner'
    }
    url = urls.reverse('gitweb:edit',
                       kwargs={'name': EXISTING_REPOS[0]['name']})
    request = rf.post(url, data=form_data)
    view = views.EditRepoView.as_view()
    response, _ = make_request(request, view, **EXISTING_REPOS[0])

    assert response.context_data['form'].errors['name'][
        0] == 'Invalid repository name.'
    assert response.status_code == 200


def test_edit_repository_no_change_view(rf):
    """Test that not changing any values don't edit the repo."""
    with patch('plinth.modules.gitweb.edit_repo') as edit_repo:
        form_data = {
            'gitweb-name': EXISTING_REPOS[0]['name'],
            'gitweb-description': EXISTING_REPOS[0]['description'],
            'gitweb-owner': EXISTING_REPOS[0]['owner'],
            'gitweb-default_branch': EXISTING_REPOS[0]['default_branch'],
        }
        request = rf.post(
            urls.reverse('gitweb:edit',
                         kwargs={'name': EXISTING_REPOS[0]['name']}),
            data=form_data)
        view = views.EditRepoView.as_view()
        response, _ = make_request(request, view, **EXISTING_REPOS[0])

        assert not edit_repo.called, 'method should not have been called'
        assert response.status_code == 302


def test_edit_repository_failed_view(rf):
    """Test that failed repo editing sends correct error message."""
    with patch('plinth.modules.gitweb.edit_repo',
               side_effect=PermissionError('Error')):
        form_data = {
            'gitweb-name': 'something_other',
            'gitweb-description': 'test-description',
            'gitweb-owner': 'test-owner',
            'gitweb-default_branch': 'main',
        }
        request = rf.post(
            urls.reverse('gitweb:edit',
                         kwargs={'name': EXISTING_REPOS[0]['name']}),
            data=form_data)
        view = views.EditRepoView.as_view()
        response, messages = make_request(request, view, **EXISTING_REPOS[0])

        assert list(
            messages)[0].message == 'An error occurred during configuration.'
        assert response.status_code == 302


def test_delete_repository_confirmation_view(rf):
    """Test that repo deletion confirmation shows correct repo name."""
    response, _ = make_request(rf.get(''), views.delete,
                               name=EXISTING_REPOS[0]['name'])

    assert response.context_data['name'] == EXISTING_REPOS[0]['name']
    assert response.status_code == 200


def test_delete_repository_view(rf):
    """Test that repo deletion sends correct success message."""
    response, messages = make_request(rf.post(''), views.delete,
                                      name=EXISTING_REPOS[0]['name'])

    assert list(messages)[0].message == '{} deleted.'.format(
        EXISTING_REPOS[0]['name'])
    assert response.status_code == 302


def test_delete_repository_fail_view(rf):
    """Test that failed repository deletion sends correct error message."""

    with patch('plinth.modules.gitweb.privileged.delete_repo',
               side_effect=FileNotFoundError('Error')):
        response, messages = make_request(rf.post(''), views.delete,
                                          name=EXISTING_REPOS[0]['name'])

        assert list(
            messages)[0].message == 'Could not delete {}: Error'.format(
                EXISTING_REPOS[0]['name'])
        assert response.status_code == 302


def test_delete_non_existing_repository_view(rf):
    """Test that deleting non-existing repository raises 404."""
    with pytest.raises(Http404):
        make_request(rf.post(''), views.delete, name='unknown_repo')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for api for android app.
"""

import copy
import json

from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse
from django.templatetags.static import static

from plinth import frontpage


def shortcuts(request, **kwargs):
    """API view to return the list of frontpage services."""
    # XXX: Get the module (or module name) from shortcut properly.
    username = str(request.user) if request.user.is_authenticated else None
    response = get_shortcuts_as_json(username)
    return HttpResponse(json.dumps(response, cls=DjangoJSONEncoder),
                        content_type='application/json')


def get_shortcuts_as_json(username=None):
    shortcuts = [
        _get_shortcut_data(shortcut)
        for shortcut in frontpage.Shortcut.list(username)
        if shortcut.component_id and shortcut.is_enabled()
    ]
    shortcuts += frontpage.get_custom_shortcuts()['shortcuts']
    return {'shortcuts': shortcuts}


def _get_shortcut_data(shortcut):
    """Return detailed information about a shortcut."""
    shortcut_data = {
        'name': shortcut.name,
        'description': shortcut.description,
        'icon_url': _get_icon_url(shortcut.app_id, shortcut.icon),
        'clients': copy.deepcopy(shortcut.clients),
        'tags': copy.deepcopy(shortcut.tags),
    }
    # XXX: Fix the hardcoding
    if shortcut.name.startswith('shortcut-ikiwiki-'):
        shortcut_data['clients'][0]['platforms'][0]['url'] += '/{}'.format(
            shortcut['name'])
    return shortcut_data


def _get_icon_url(app_id, icon_name):
    """Return icon path given app's ID and icon name."""
    if not icon_name:
        return None

    if icon_name.startswith('/'):
        return icon_name

    if app_id:
        return static(f'{app_id}') + f'/icons/{icon_name}.png'

    return static(f'theme/icons/{icon_name}.png')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Utilities for parsing dbconfig-common files with Augeas."""

import pathlib

import augeas


def get_credentials(dbconfig_path: str) -> dict[str, str]:
    """Parse dbconfig-common file with Augeas Shellvars lens."""
    if not pathlib.Path(dbconfig_path).is_file():
        raise FileNotFoundError(f'DB config not found: {dbconfig_path}')

    aug = _load_augeas(dbconfig_path)

    required = ['dbc_dbuser', 'dbc_dbpass', 'dbc_dbname']
    credentials = {}
    for key in required + ['dbc_dbserver']:
        credentials[key] = aug.get(key).strip('\'"')

    if not all(credentials.get(key) for key in required):
        raise ValueError('Missing required dbconfig-common credentials')

    return {
        'user': credentials['dbc_dbuser'],
        'password': credentials['dbc_dbpass'],
        'database': credentials['dbc_dbname'],
        'host': credentials['dbc_dbserver'] or 'localhost'
    }


def _load_augeas(config_path: str):
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    pathstr = str(config_path)
    aug.transform('Shellvars', pathstr)
    aug.set('/augeas/context', f'/files{pathstr}')
    aug.load()
    return aug



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure Coturn server."""

import logging
import pathlib
from typing import Iterator

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import menu
from plinth.daemon import Daemon
from plinth.modules import names
from plinth.modules.backups.components import BackupRestore
from plinth.modules.coturn.components import TurnConfiguration, TurnConsumer
from plinth.modules.firewall.components import Firewall
from plinth.modules.letsencrypt.components import LetsEncrypt
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    _('Coturn is a server to facilitate audio/video calls and conferences by '
      'providing an implementation of TURN and STUN protocols. WebRTC, SIP '
      'and other communication servers can use it to establish a call between '
      'parties who are otherwise unable connect to each other.'),
    format_lazy(
        _('It is not meant to be used directly by users. Servers such as '
          '<a href="{ms_url}">Matrix Synapse</a> or <a href="{e_url}">ejabberd'
          '</a> need to be configured with the details provided here.'),
        ms_url=reverse_lazy('matrixsynapse:index'),
        e_url=reverse_lazy('ejabberd:index')),
]

logger = logging.getLogger(__name__)


class CoturnApp(app_module.App):
    """FreedomBox app for Coturn."""

    app_id = 'coturn'

    _version = 2

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('Coturn'), icon_filename='coturn',
                               description=_description, manual_page='Coturn',
                               tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-coturn', info.name, info.icon_filename,
                              info.tags, 'coturn:index',
                              parent_url_name='apps')
        self.add(menu_item)

        # Include sqlite3 to prevent removal of coturn from removal of
        # roundcube.
        packages = Packages('packages-coturn', ['coturn', 'sqlite3'])
        self.add(packages)

        firewall = Firewall('firewall-coturn', info.name,
                            ports=['coturn-freedombox'], is_external=True)
        self.add(firewall)

        letsencrypt = LetsEncrypt(
            'letsencrypt-coturn', domains=get_domains, daemons=['coturn'],
            should_copy_certificates=True,
            private_key_path='/etc/coturn/certs/pkey.pem',
            certificate_path='/etc/coturn/certs/cert.pem',
            user_owner='turnserver', group_owner='turnserver',
            managing_app='coturn')
        self.add(letsencrypt)

        daemon = Daemon(
            'daemon-coturn', 'coturn', listen_ports=[(3478, 'udp4'),
                                                     (3478, 'udp6'),
                                                     (3478, 'tcp4'),
                                                     (3478, 'tcp6'),
                                                     (3479, 'udp4'),
                                                     (3479, 'udp6'),
                                                     (3479, 'tcp4'),
                                                     (3479, 'tcp6'),
                                                     (5349, 'udp4'),
                                                     (5349, 'udp6'),
                                                     (5349, 'tcp4'),
                                                     (5349, 'tcp6'),
                                                     (5350, 'udp4'),
                                                     (5350, 'udp6'),
                                                     (5350, 'tcp4'),
                                                     (5350, 'tcp6')])
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-coturn',
                                          reserved_usernames=['turnserver'])
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-coturn',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if old_version == 0:
            self.enable()

        self.get_component('letsencrypt-coturn').setup_certificates()
        notify_configuration_change()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()


def get_available_domains() -> Iterator[str]:
    """Return an iterator with all domains able to have a certificate."""
    return (domain.name for domain in names.components.DomainName.list()
            if domain.domain_type.can_have_certificate)


def get_domain():
    """Read TLS domain from config file select first available if none."""
    config = get_config()
    if config.domain:
        return config.domain

    domain = next(get_available_domains(), None)
    set_domain(domain)

    return domain


def get_domains():
    """Return a list with the configured domains."""
    # If not installed, return empty. But work while installing too.
    if not pathlib.Path('/etc/coturn/freedombox.conf').exists():
        return []

    domain = get_domain()
    if domain:
        return [domain]

    return []


def set_domain(domain):
    """Set the TLS domain by writing a file to data directory."""
    if domain:
        privileged.set_domain(domain)
        notify_configuration_change()


def get_config():
    """Return the coturn server configuration."""
    config = privileged.get_config()
    return TurnConfiguration(config['realm'], [], config['static_auth_secret'])


def notify_configuration_change():
    """Notify all coturn components about the new configuration."""
    logger.info('Notifying STUN/TURN consumers about configuration change')
    config = get_config()
    for component in TurnConsumer.list():
        component.on_config_change(config)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Simple key/value store using Django models
"""

from . import db


def get(key):
    """Return the value of a key"""
    from plinth.models import KVStore

    with db.lock:
        # pylint: disable-msg=E1101
        return KVStore.objects.get(pk=key).value


def get_default(key, default_value):
    """Return the value of the key if key exists else return default_value"""
    with db.lock:
        try:
            return get(key)
        except Exception:
            return default_value


def set(key, value):  # pylint: disable-msg=W0622
    """Store the value of a key"""
    from plinth.models import KVStore
    with db.lock:
        store = KVStore(key=key, value=value)
        store.save()


def delete(key, ignore_missing=False):
    """Delete a key"""
    from plinth.models import KVStore
    with db.lock:
        try:
            return KVStore.objects.get(key=key).delete()
        except KVStore.DoesNotExist:
            if not ignore_missing:
                raise



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for Kiwix views.
"""

from unittest.mock import MagicMock, call, patch

import pytest
from django import urls
from django.contrib.messages.storage.fallback import FallbackStorage
from django.core.files.uploadedfile import SimpleUploadedFile
from django.http.response import Http404

from plinth import module_loader
from plinth.modules.kiwix import views

# For all tests, use plinth.urls instead of urls configured for testing
pytestmark = pytest.mark.urls('plinth.urls')

ZIM_ID = 'bc4f8cdf-5626-2b13-3860-0033deddfbea'


@pytest.fixture(autouse=True, scope='module')
def fixture_kiwix_urls():
    """Make sure kiwix app's URLs are part of plinth.urls."""
    with patch('plinth.module_loader._modules_to_load', new=[]) as modules, \
            patch('plinth.urls.urlpatterns', new=[]):
        modules.append('plinth.modules.kiwix')
        module_loader.include_urls()
        yield


def make_request(request, view, **kwargs):
    """Make request with a message storage."""
    setattr(request, 'session', 'session')
    messages = FallbackStorage(request)
    setattr(request, '_messages', messages)
    response = view(request, **kwargs)

    return response, messages


@pytest.fixture(autouse=True)
def fixture_kiwix_patch():
    """Patch kiwix methods."""
    with patch(
            'plinth.modules.kiwix.privileged.list_packages') as list_libraries:
        list_libraries.return_value = {
            ZIM_ID: {
                'title': 'TestExistingPackage',
                'description': 'A sample content package',
                'path': 'test_existing_package'
            }
        }
        yield


@pytest.fixture()
def file_path(tmp_path):
    return str(tmp_path / 'FreedomBox.zim')


def uploaded_file():
    return SimpleUploadedFile('FreedomBox.zim', content=b'FreedomBox rocks!',
                              content_type='application/octet-stream')


@pytest.fixture()
def add_package_request(rf, file_path):
    """Patch add_package."""
    post_data = {'kiwix-file': uploaded_file()}
    request = rf.post('', data=post_data)
    request.FILES['kiwix-file'].temporary_file_path = MagicMock(
        return_value=file_path)
    return request


@patch('plinth.modules.kiwix.privileged.add_package')
def test_add_package(add_package, file_path, add_package_request):
    """Test that adding content view works."""
    response, messages = make_request(add_package_request,
                                      views.AddPackageView.as_view())
    assert response.status_code == 302
    assert response.url == urls.reverse('kiwix:index')
    assert list(messages)[0].message == 'Content package added.'
    add_package.assert_has_calls([call('FreedomBox.zim', file_path)])


@patch('plinth.modules.kiwix.privileged.add_package')
def test_add_package_failed(add_package, add_package_request):
    """Test that adding content package fails in case of an error."""
    add_package.side_effect = RuntimeError('TestError')
    response, messages = make_request(add_package_request,
                                      views.AddPackageView.as_view())
    assert response.status_code == 302
    assert response.url == urls.reverse('kiwix:index')
    assert list(messages)[0].message.startswith(
        'Failed to add content package.')


@patch('plinth.app.App.get')
def test_delete_package_confirmation_view(_app, rf):
    """Test that deleting content confirmation shows correct title."""
    response, _ = make_request(rf.get(''), views.delete_package, zim_id=ZIM_ID)
    assert response.status_code == 200
    assert response.context_data['name'] == 'TestExistingPackage'


@patch('plinth.modules.kiwix.privileged.delete_package')
@patch('plinth.app.App.get')
def test_delete_package(_app, delete_package, rf):
    """Test that deleting a content package works."""
    response, messages = make_request(rf.post(''), views.delete_package,
                                      zim_id=ZIM_ID)
    assert response.status_code == 302
    assert response.url == urls.reverse('kiwix:index')
    assert list(messages)[0].message == 'TestExistingPackage deleted.'
    delete_package.assert_has_calls([call(ZIM_ID)])


@patch('plinth.modules.kiwix.privileged.delete_package')
def test_delete_package_error(delete_package, rf):
    """Test that deleting content shows an error when operation fails."""
    delete_package.side_effect = ValueError('TestError')
    response, messages = make_request(rf.post(''), views.delete_package,
                                      zim_id=ZIM_ID)
    assert response.status_code == 302
    assert response.url == urls.reverse('kiwix:index')
    assert list(messages)[0].message == \
        'Could not delete TestExistingPackage: TestError'


def test_delete_package_non_existing(rf):
    """Test that deleting content shows error when operation fails."""
    with pytest.raises(Http404):
        make_request(rf.post(''), views.delete_package,
                     zim_id='NonExistentZimId')

    with pytest.raises(Http404):
        make_request(rf.get(''), views.delete_package,
                     zim_id='NonExistentZimId')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure roundcube."""

import pathlib
import re

from plinth import action_utils
from plinth.actions import privileged

_config_file = pathlib.Path('/etc/roundcube/freedombox-config.php')
_rc_db_file = pathlib.Path('/var/lib/dbconfig-common/sqlite3/'
                           'roundcube/roundcube')


@privileged
def pre_install():
    """Preseed debconf values before packages are installed."""
    action_utils.debconf_set_selections([
        'roundcube-core roundcube/dbconfig-install boolean true',
        'roundcube-core roundcube/database-type string sqlite3'
    ])


@privileged
def setup():
    """Add FreedomBox configuration and include from main configuration."""
    if not _config_file.exists():
        _config_file.write_text('<?php\n', encoding='utf-8')

    base_config = pathlib.Path('/etc/roundcube/config.inc.php')
    lines = base_config.read_text(encoding='utf-8').splitlines()
    exists = any((str(_config_file) in line for line in lines))
    if not exists:
        lines.append(f'include_once("{_config_file}");\n')
        base_config.write_text('\n'.join(lines), encoding='utf-8')


@privileged
def get_config() -> dict[str, bool]:
    """Return the current configuration as a dictionary."""
    pattern = r'\s*\$config\[\s*\'([^\']*)\'\s*\]\s*=\s*\'([^\']*)\'\s*;'
    _config = {}
    try:
        for line in _config_file.read_text(encoding='utf-8').splitlines():
            match = re.match(pattern, line)
            if match:
                _config[match.group(1)] = match.group(2)
    except FileNotFoundError:
        pass

    local_only = _config.get('default_host') == 'localhost'
    return {'local_only': local_only}


@privileged
def set_config(local_only: bool):
    """Set the configuration."""
    config = '''<?php
$config['log_driver'] = 'syslog';
'''

    if local_only:
        config += '''
$config['default_host'] = 'localhost';
$config['mail_domain'] = '%n';

$config['smtp_server'] = 'localhost';
$config['smtp_port'] = 25;
$config['smtp_helo_host'] = 'localhost';
'''

    _config_file.write_text(config, encoding='utf-8')


@privileged
def uninstall():
    """Remove config file and database."""
    for item in _config_file, _rc_db_file:
        item.unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Django views for Home Assistant app."""

import logging

from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from plinth.forms import DomainSelectionForm
from plinth.views import AppView

logger = logging.getLogger(__name__)


class HomeAssistantAppView(AppView):
    """Show Home Assistant app main view."""

    app_id = 'homeassistant'
    template_name = 'homeassistant.html'
    form_class = DomainSelectionForm

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        component = self.app.get_component('webserverroot-homeassistant')
        initial.update({
            'domain_name': component.domain_get() or '',
        })
        return initial

    def get_form_kwargs(self):
        """Return the arguments to instantiate form with."""
        kwargs = super().get_form_kwargs()
        kwargs['show_none'] = True
        return kwargs

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        old_config = self.get_initial()
        new_config = form.cleaned_data

        is_changed = False

        def _value_changed(key):
            return old_config.get(key) != new_config.get(key)

        if _value_changed('domain_name'):
            component = self.app.get_component('webserverroot-homeassistant')
            component.domain_set(new_config['domain_name'] or None)
            is_changed = True

        if is_changed:
            messages.success(self.request, _('Configuration updated.'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Views for the Kiwix module.
"""

import logging

from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.http import Http404
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.urls import reverse_lazy
from django.utils.translation import gettext as _
from django.views.generic.edit import FormView

from plinth import app as app_module
from plinth import views
from plinth.errors import PlinthError
from plinth.modules import storage
from plinth.modules.kiwix import privileged
from plinth.views import messages_error

from . import forms

logger = logging.getLogger(__name__)


class KiwixAppView(views.AppView):
    """Serve configuration form."""

    app_id = 'kiwix'
    template_name = 'kiwix.html'

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['packages'] = privileged.list_packages()
        return context


class AddPackageView(SuccessMessageMixin, FormView):
    """View to add content package in the form of ZIM files."""

    form_class = forms.AddPackageForm
    prefix = 'kiwix'
    template_name = 'kiwix-add-package.html'
    success_url = reverse_lazy('kiwix:index')
    success_message = _('Content package added.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Add a new content package')

        # TODO The following is almost duplicated in backups/views.py
        try:
            mount_info = storage.get_mount_info('/')
        except PlinthError as exception:
            logger.exception(
                'Error getting information about root partition: %s',
                exception)
        else:
            context['max_filesize'] = storage.format_bytes(
                mount_info['free_bytes'])

        return context

    def form_valid(self, form):
        """Store the uploaded file."""
        try:
            uploaded_file = self.request.FILES['kiwix-file']
            privileged.add_package(uploaded_file.name,
                                   uploaded_file.temporary_file_path())
        except FileExistsError:
            messages.error(self.request, _('Content package already exists.'))
            return redirect(reverse_lazy('kiwix:index'))
        except Exception as exception:
            messages_error(self.request, _('Failed to add content package.'),
                           exception)
            return redirect(reverse_lazy('kiwix:index'))

        return super().form_valid(form)


def delete_package(request, zim_id):
    """View to delete a library."""
    packages = privileged.list_packages()
    if zim_id not in packages:
        raise Http404

    name = packages[zim_id]['title']

    if request.method == 'POST':
        try:
            privileged.delete_package(zim_id)
            messages.success(request, _('{name} deleted.').format(name=name))
        except Exception as error:
            messages.error(
                request,
                _('Could not delete {name}: {error}').format(
                    name=name, error=error))
        return redirect(reverse_lazy('kiwix:index'))

    return TemplateResponse(request, 'kiwix-delete-package.html', {
        'title': app_module.App.get('kiwix').info.name,
        'name': name
    })



# SPDX-License-Identifier: AGPL-3.0-or-later

#
# Generated by Django 1.9 on 2015-12-04 07:27
#
"""
Initial Django migration for FreedomBox to create database tables.
"""

from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    """Initial migration schema for FreedomBox models."""

    initial = True

    dependencies: list = []

    operations = [
        migrations.CreateModel(
            name='KVStore',
            fields=[
                ('key', models.TextField(primary_key=True, serialize=False)),
                ('value_json', models.TextField()),
            ],
        ),
    ]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure backups (with borg) and sshfs."""

import functools
import json
import os
import pathlib
import re
import shutil
import subprocess
import tarfile

from django.utils.translation import gettext_lazy as _

from plinth import action_utils, actions
from plinth import app as app_module
from plinth import module_loader
from plinth.actions import privileged, secret_str
from plinth.utils import Version

from . import errors

TIMEOUT = 30
BACKUPS_DATA_PATH = pathlib.Path('/var/lib/plinth/backups-data/')
BACKUPS_UPLOAD_PATH = pathlib.Path('/var/lib/freedombox/backups-upload/')
MANIFESTS_FOLDER = '/var/lib/plinth/backups-manifests/'

# known errors that come up when remotely accessing a borg repository
# 'errors' are error strings to look for in the stacktrace.
KNOWN_ERRORS = [
    {
        'errors': ['subprocess.TimeoutExpired'],
        'message':
            _('Connection refused - make sure you provided correct '
              'credentials and the server is running.'),
        'raise_as':
            errors.BorgError,
    },
    {
        'errors': ['Connection refused'],
        'message': _('Connection refused'),
        'raise_as': errors.BorgError,
    },
    {
        'errors': [
            'not a valid repository', 'does not exist', 'FileNotFoundError'
        ],
        'message': _('Repository not found'),
        'raise_as': errors.BorgRepositoryDoesNotExistError,
    },
    {
        'errors': ['[Pp]assphrase supplied in .* is incorrect'],
        'message': _('Incorrect encryption passphrase'),
        'raise_as': errors.BorgError,
    },
    {
        'errors': ['Connection reset by peer'],
        'message': _('SSH access denied'),
        'raise_as': errors.SshfsError,
    },
    {
        'errors': ['There is already something at'],
        'message':
            _('Repository path is neither empty nor '
              'is an existing backups repository.'),
        'raise_as':
            errors.BorgError,
    },
    {
        'errors': ['A repository already exists at'],
        'message': None,
        'raise_as': errors.BorgRepositoryExists,
    },
    {
        'errors': ['Archive .* already exists'],
        'message':
            _('An archive with given name already exists in the repository.'),
        'raise_as':
            errors.BorgArchiveExists,
    },
    {
        'errors': ['Archive .* not found'],
        'message':
            _('Archive with given name was not found in the repository.'),
        'raise_as':
            errors.BorgArchiveDoesNotExist,
    },
    {
        'errors': ['Failed to create/acquire the lock'],
        'message': _('Backup system is busy with another operation.'),
        'raise_as': errors.BorgBusy,
    },
    {
        'errors': ['No space left on device'],
        'message': _('Not enough space left on the disk or remote location.'),
        'raise_as': errors.BorgNoSpace,
    },
]


class AlreadyMountedError(Exception):
    """Exception raised when mount point is already mounted."""


def reraise_known_errors(privileged_func):
    """Decorator to convert borg raised exceptions to specialized ones."""

    @functools.wraps(privileged_func)
    def wrapper(*args, **kwargs):
        """Run privileged method, catch exceptions and throw new ones."""
        try:
            return privileged_func(*args, **kwargs)
        except Exception as exception:
            _reraise_known_errors(exception)

    return wrapper


def _reraise_known_errors(err):
    """Look whether the caught error is known and reraise it accordingly"""
    stdout = (getattr(err, 'stdout', b'') or b'').decode()
    stderr = (getattr(err, 'stderr', b'') or b'').decode()
    caught_error = str((err, err.args, stdout, stderr))
    for known_error in KNOWN_ERRORS:
        for error in known_error['errors']:
            if re.search(error, caught_error):
                raise known_error['raise_as'](known_error['message'])

    raise err


@reraise_known_errors
@privileged
def mount(mountpoint: str, remote_path: str, ssh_keyfile: str,
          user_known_hosts_file: str = '/dev/null'):
    """Mount a remote ssh path via sshfs."""
    try:
        _validate_mountpoint(mountpoint)
    except AlreadyMountedError:
        return

    input_ = None
    # the shell would expand ~/ to the local home directory
    remote_path = remote_path.replace('~/', '').replace('~', '')
    # 'reconnect', 'ServerAliveInternal' and 'ServerAliveCountMax' allow the
    # client (FreedomBox) to keep control of the SSH connection even when the
    # SSH server misbehaves. Without these options, other commands such as
    # 'freedombox-cmd storage usage_info --no-args', 'df', 'freedombox-cmd
    # sshfs is_mounted --no-args', or 'mountpoint' might block indefinitely
    # (even when manually invoked from the command line). This situation has
    # some lateral effects, causing major system instability in the course of
    # ~11 days, and leaving the system in such state that the only solution is
    # a reboot.
    cmd = [
        'sshfs', remote_path, mountpoint, '-o',
        f'UserKnownHostsFile={user_known_hosts_file}', '-o',
        'StrictHostKeyChecking=yes', '-o', 'reconnect', '-o',
        'ServerAliveInterval=15', '-o', 'ServerAliveCountMax=3', '-o',
        'IdentityFile=' + ssh_keyfile
    ]
    action_utils.run(cmd, check=True, timeout=TIMEOUT, input=input_)


@reraise_known_errors
@privileged
def umount(mountpoint: str):
    """Unmount a mountpoint."""
    action_utils.run(['umount', mountpoint], check=True)


def _validate_mountpoint(mountpoint):
    """Check that the folder is empty, and create it if it doesn't exist."""
    if os.path.exists(mountpoint):
        if _is_mounted(mountpoint):
            raise AlreadyMountedError('Mountpoint %s already mounted' %
                                      mountpoint)
        if os.listdir(mountpoint) or not os.path.isdir(mountpoint):
            raise ValueError('Mountpoint %s is not an empty directory' %
                             mountpoint)
    else:
        os.makedirs(mountpoint)


def _is_mounted(mountpoint):
    """Return boolean whether a local directory is a mountpoint."""
    cmd = ['mountpoint', '-q', mountpoint]
    # mountpoint exits with status non-zero if it didn't find a mountpoint
    try:
        action_utils.run(cmd, check=True)
        return True
    except subprocess.CalledProcessError:
        return False


@reraise_known_errors
@privileged
def is_mounted(mount_point: str) -> bool:
    """Return whether a path is already mounted."""
    return _is_mounted(mount_point)


@reraise_known_errors
@privileged
def setup(path: str):
    """Create repository if it does not already exist."""
    _create_backup_data_directory()

    try:
        _run(['borg', 'info', path], check=True)
    except subprocess.CalledProcessError:
        parent = os.path.dirname(path)
        if not os.path.exists(parent):
            os.makedirs(parent)

        _init_repository(path, encryption='none')


def _create_backup_data_directory():
    """Create the backups-data with proper permissions."""
    old_umask = os.umask(0o077)
    try:
        BACKUPS_DATA_PATH.mkdir(exist_ok=True)
        BACKUPS_DATA_PATH.chmod(0o700)
        shutil.chown(BACKUPS_DATA_PATH, 'root', 'root')
    finally:
        os.umask(old_umask)


def _init_repository(path: str, encryption: str,
                     encryption_passphrase: str | None = None):
    """Initialize a local or remote borg repository."""
    if encryption != 'none':
        if not encryption_passphrase:
            raise ValueError('No encryption passphrase provided')

    cmd = ['borg', 'init', '--encryption', encryption, path]
    _run(cmd, encryption_passphrase)


@reraise_known_errors
@privileged
def init(path: str, encryption: str,
         encryption_passphrase: secret_str | None = None):
    """Initialize the borg repository."""
    _init_repository(path, encryption, encryption_passphrase)


@reraise_known_errors
@privileged
def info(path: str, encryption_passphrase: secret_str | None = None) -> dict:
    """Show repository information."""
    process = _run(['borg', 'info', '--json', path], encryption_passphrase)
    return json.loads(process.stdout.decode())


@reraise_known_errors
@privileged
def list_repo(path: str,
              encryption_passphrase: secret_str | None = None) -> dict:
    """List repository contents."""
    process = _run(['borg', 'list', '--json', '--format="{comment}"', path],
                   encryption_passphrase)
    return json.loads(process.stdout.decode())


@reraise_known_errors
@privileged
def add_uploaded_archive(file_name: str, temporary_file_path: str):
    """Store an archive uploaded by the user."""
    BACKUPS_UPLOAD_PATH.mkdir(parents=True, exist_ok=True)
    action_utils.move_uploaded_file(temporary_file_path, BACKUPS_UPLOAD_PATH,
                                    file_name, allow_overwrite=True,
                                    permissions=0o600)


@reraise_known_errors
@privileged
def remove_uploaded_archive(file_path: str):
    """Delete the archive uploaded by the user."""
    resolved_file_path = pathlib.Path(file_path).resolve()
    if (resolved_file_path.is_relative_to(BACKUPS_UPLOAD_PATH)
            and resolved_file_path.is_file()):
        resolved_file_path.unlink()


def _get_borg_version():
    """Return the version of borgbackup."""
    process = _run(['borg', '--version'])
    return process.stdout.decode().split()[1]  # Example: "borg 1.1.9"


@reraise_known_errors
@privileged
def create_archive(path: str, paths: list[str], comment: str | None = None,
                   encryption_passphrase: secret_str | None = None):
    """Create archive."""
    existing_paths = filter(os.path.exists, paths)
    command = ['borg', 'create', '--json']
    if comment:
        if Version(_get_borg_version()) < Version('1.1.10'):
            # Undo any placeholder escape sequences in comments as this version
            # of borg does not support placeholders. XXX: Drop this code when
            # support for borg < 1.1.10 is dropped.
            comment = comment.replace('{{', '{').replace('}}', '}')

        command += ['--comment', comment]

    command += [path] + list(existing_paths)
    _run(command, encryption_passphrase)


@reraise_known_errors
@privileged
def delete_archive(path: str, encryption_passphrase: secret_str | None = None):
    """Delete archive."""
    _run(['borg', 'delete', path], encryption_passphrase)


def _extract(archive_path, destination, encryption_passphrase, locations=None):
    """Extract archive contents."""
    prev_dir = os.getcwd()
    borg_call = ['borg', 'extract', archive_path]
    # do not extract any files when we get an empty locations list
    if locations is not None:
        borg_call.extend(locations)

    try:
        os.chdir(os.path.expanduser(destination))
        process = _run(borg_call, encryption_passphrase, check=False)
        if process.returncode != 0:
            error = process.stderr.decode()
            # Don't fail on the borg error when no files were matched
            if "never matched" not in error:
                raise subprocess.CalledProcessError(process.returncode,
                                                    process.args)
    finally:
        os.chdir(prev_dir)


@reraise_known_errors
@privileged
def export_tar(path: str, encryption_passphrase: secret_str | None = None):
    """Export archive contents as tar stream on stdout."""
    env = _get_env(encryption_passphrase)
    process = subprocess.Popen(
        ['borg', 'export-tar', path, '-', '--tar-filter=gzip'], env=env,
        stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    return actions.ProcessBufferedReader(process)


def _read_archive_file(archive, filepath, encryption_passphrase):
    """Read the content of a file inside an archive."""
    borg_call = ['borg', 'extract', archive, filepath, '--stdout']
    return _run(borg_call, encryption_passphrase).stdout.decode()


@reraise_known_errors
@privileged
def get_archive_apps(
        path: str,
        encryption_passphrase: secret_str | None = None) -> list[str]:
    """Get list of apps included in archive."""
    manifest_folder = os.path.relpath(MANIFESTS_FOLDER, '/')
    borg_call = [
        'borg', 'list', path, manifest_folder, '--format', '{path}{NEWLINE}'
    ]
    try:
        borg_process = _run(borg_call, encryption_passphrase)
        manifest_path = borg_process.stdout.decode().strip()
    except subprocess.CalledProcessError:
        raise RuntimeError('Borg exited unsuccessfully')

    manifest = None
    if manifest_path:
        manifest_data = _read_archive_file(path, manifest_path,
                                           encryption_passphrase)
        manifest = json.loads(manifest_data)

    archive_apps = []
    if manifest:
        for app in _get_apps_of_manifest(manifest):
            archive_apps.append(app['name'])

    return archive_apps


def _get_apps_of_manifest(manifest):
    """Get apps of a manifest.

    Supports both dict format as well as list format of plinth <=0.42

    """
    if isinstance(manifest, list):
        apps = manifest
    elif isinstance(manifest, dict) and 'apps' in manifest:
        apps = manifest['apps']
    else:
        raise RuntimeError('Unknown manifest format')

    return apps


@reraise_known_errors
@privileged
def get_exported_archive_apps(path: str) -> list[str]:
    """Get list of apps included in an exported archive file."""
    manifest = None
    with tarfile.open(path) as tar_handle:
        filenames = tar_handle.getnames()
        for name in filenames:
            if 'var/lib/plinth/backups-manifests/' in name \
               and name.endswith('.json'):
                file_handle = tar_handle.extractfile(name)
                if not file_handle:
                    raise RuntimeError(
                        'Unable to extract app manifest from backup file.')

                manifest_data = file_handle.read()
                manifest = json.loads(manifest_data)
                break

    app_names = []
    if manifest:
        for app in _get_apps_of_manifest(manifest):
            app_names.append(app['name'])

    return app_names


@reraise_known_errors
@privileged
def restore_archive(archive_path: str, destination: str,
                    directories: list[str], files: list[str],
                    encryption_passphrase: secret_str | None = None):
    """Restore files from an archive."""
    locations_all = directories + files
    locations_all = [
        os.path.relpath(location, '/') for location in locations_all
    ]
    _extract(archive_path, destination, encryption_passphrase,
             locations=locations_all)


@reraise_known_errors
@privileged
def restore_exported_archive(path: str, directories: list[str],
                             files: list[str]):
    """Restore files from an exported archive."""
    with tarfile.open(path) as tar_handle:
        for member in tar_handle.getmembers():
            path = '/' + member.name
            if path in files:
                tar_handle.extract(member, '/')
            else:
                for directory in directories:
                    if path.startswith(directory):
                        tar_handle.extract(member, '/')
                        break


def _assert_app_id(app_id):
    """Check that app ID is correct."""
    if not re.fullmatch(r'[a-z0-9_]+', app_id):
        raise Exception('Invalid App ID')


@reraise_known_errors
@privileged
def dump_settings(app_id: str, settings: dict[str, int | float | bool | str]):
    """Dump an app's settings to a JSON file."""
    _assert_app_id(app_id)
    settings_path = BACKUPS_DATA_PATH / f'{app_id}-settings.json'
    settings_path.write_text(json.dumps(settings))


@reraise_known_errors
@privileged
def load_settings(app_id: str) -> dict[str, int | float | bool | str]:
    """Load an app's settings from a JSON file."""
    _assert_app_id(app_id)
    settings_path = BACKUPS_DATA_PATH / f'{app_id}-settings.json'
    try:
        return json.loads(settings_path.read_text())
    except FileNotFoundError:
        return {}


@privileged
def delete_before_restore(app_id: str):
    """Delete some paths before restoring an app."""
    module_loader.load_modules()
    app_module.apps_init()
    app = app_module.App.get(app_id)

    from plinth.modules.backups.components import BackupRestore
    components = app.get_components_of_type(BackupRestore)
    for component in components:
        for path in component.delete_before_restore:
            shutil.rmtree(path, ignore_errors=True)


def _get_env(encryption_passphrase: str | None = None):
    """Create encryption and ssh kwargs out of given arguments."""
    env = dict(os.environ, BORG_RELOCATED_REPO_ACCESS_IS_OK='yes',
               BORG_UNKNOWN_UNENCRYPTED_REPO_ACCESS_IS_OK='yes',
               LANG='C.UTF-8')
    # Always provide BORG_PASSPHRASE (also if empty) so borg does not get stuck
    # while asking for a passphrase.
    env['BORG_PASSPHRASE'] = encryption_passphrase or ''
    return env


def _run(cmd, encryption_passphrase=None, check=True, **kwargs):
    """Wrap the command with extra encryption passphrase handling."""
    env = _get_env(encryption_passphrase)
    return action_utils.run(cmd, check=check, env=env, **kwargs)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for Tor Proxy module.
"""

from unittest.mock import patch

import pytest

from plinth.modules.torproxy import utils


class TestTorProxy:
    """Test cases for testing the Tor Proxy module."""

    @staticmethod
    @pytest.mark.usefixtures('needs_root')
    def test_is_apt_transport_tor_enabled():
        """Test that is_apt_transport_tor_enabled does not raise any unhandled
        exceptions.
        """
        utils.is_apt_transport_tor_enabled()

    @staticmethod
    @patch('plinth.app.App.get')
    @pytest.mark.usefixtures('needs_root', 'load_cfg')
    def test_get_status(_app_get):
        """Test that get_status does not raise any unhandled exceptions.

        This should work regardless of whether tor is installed, or
        /etc/tor/instances/fbxproxy/torrc exists.
        """
        utils.get_status()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Django form for configuring calibre."""

from django import forms
from django.core import validators
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from . import privileged


class CreateLibraryForm(forms.Form):
    """Form to create an empty library."""

    name = forms.CharField(
        label=_('Name of the new library'), strip=True,
        help_text=_('Only letters of the English alphabet, numbers '
                    'and the characters _ . and - without spaces or '
                    'special characters. Example: My_Library_2000'),
        validators=[validators.RegexValidator(r'^[A-Za-z0-9_.-]+$')])

    def clean_name(self):
        """Check if the library name is valid."""
        name = self.cleaned_data['name']

        if name in privileged.list_libraries():
            raise ValidationError(
                _('A library with this name already exists.'))

        return name



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Secure Shell Server module.
"""

from django.urls import re_path

from plinth.modules.ssh.views import SshAppView

urlpatterns = [
    re_path(r'^sys/ssh/$', SshAppView.as_view(), name='index'),
]
